/* irred.c                                              irred version 1.35

==========================================================================
|                                                                        |
|  Copyright (C) 2000 R. P. Brent.                                       |
|                                                                        | 
|  This program is free software; you can redistribute it and/or         |
|  modify it under the terms of the GNU General Public License,          |
|  version 2, June 1991, as published by the Free Software Foundation.   |
|  For details see http://www.gnu.org/copyleft/gpl.html .                |    
|                                                                        |
|  External assembler routines (see ASM and comments below) are covered  |
|  by the same copyright conditions as this C code.                      |
|                                                                        |
==========================================================================

   Author: Richard Brent 	
      	   with assistance and contributions from
   	   Samuli Larvala and Paul Zimmermann

   Dates:  20000528..20000719
   
   Efficient test for primitivity of trinomials
   x^r + x^s +1 (mod 2) where 2^r-1 is a Mersenne prime,
   using (32-bit or 64-bit) word-length bit operations.

   Assumes 0 < s < r and r-s > 63 (no real loss of generality because
   can replace s by r-s, and small r cases are easily handled).

   NB: It is assumed that 2^r-1 is prime - this is not checked !

   Usage:
   	  irred restart-file log-file time-limit skip-file  
   	  					(all arguments optional)
   Arguments:
    
          A restart file can be specified as first command line argument.
          If no file is specified, input data is read from stdin.

            Format of restart file or stdin is one line
              r s1 s2 
            meaning to start next run at this (r,s1) and go to (r,s2-1)
            [or, if s1 > s2, go from (r,s1) to (r,s2+1)].
            If s1 = s2 then run is finished.
            
	  A log file can be specified as second command line argument.
	  If no file is specified, then log output -> stdout with an
	  identifier "LOG" (can be extracted using grep).

 	  Max minutes to run (m) can be specified as 3rd command line argument.
	
	    Maximum runtime is m minutes (infinite if m = 0).
	    
	    The parameters m and s2 may be omitted (in which case they
	    assumed to be zero).
	    
	  A skip file can be specified as the fourth command line argument.
	  This file is a list of pairs (u, v) (one per line) indicating that
	  s in the interval [u, v] should not be tested. This is useful if
	  certain intervals are already done or reserved for another machine.
	  If a skip file is desired, the parameter m must be specified
	  (possibly zero).
*/

#include <stdio.h>
#include <stdlib.h>		/* For malloc */
#include <time.h>   		/* For clock */

#define TRUE 1
#define FALSE 0
#define AND &&			/* Logical AND */
#define OR ||			/* Logical OR  */
#define NOT !			/* Logical NOT */

/* Set VERBOSE and CONTINUE to determine program behaviour */

#define VERBOSE TRUE		/* If true give more informative output */
#define CONTINUE FALSE		/* If true, continue after finding a
                                   primitive trinomial (useful for small r) */

/* Set exactly one of the following machine types to TRUE */

#define IBMPC TRUE /* 32-bit CISC, little-endian, e.g. Pentium.

		      If using the assembler routines irredasm.s,
		      compile with gcc -c irredasm.s 
		      This should generate irredasm.o 
		      (does not work for all versions of gcc). 

		      With gcc 2.95.2, compile irred.c with 
		      gcc -O1 -fomit-frame-pointer -funroll-loops
		      	-o irred irred.c irredasm.o
		      	
		      With other versions of gcc, e.g. 2.7.2.3, try
		      gcc -O1 -fomit-frame-pointer
		        -o irred irred.c irredasm.o 
	 	   */
           
#define SPARC FALSE	/* 32-bit RISC, e.g. Sun Sparc. Compile with gcc -O6 */

#define ALPHA FALSE 	/* 64-bit RISC, e.g. DEC Alpha. Compile with
			   gcc -O4 -mcpu=ev6 -Wa,-arch -Wa,ev6 -funroll-loops */

#define SGI FALSE	/* 64-bit SGI R10000. Compile with cc -64 -Ofast */
			/* Currently, SGI and ALPHA have the same effect */
			
/* Comments on timing and memory requirements:

   Sieving quickly discards about 92% of polynomials.
   Each polynomial not discarded by sieving takes time O(r^2)
   but there may be a slow-down as r increases (this is a cache effect).

   Memory requirement is 3r/8 + (constant) bytes. 
   The working set is r/4 + (constant) bytes.

32-bit version (W64 = FALSE):

   Following times are for program with ASM = TRUE (except where 
   noted - comments may refer to old flags MMX and MMX2 instead of ASM), 
   compiled with gcc version 2.7.2.3 using 
   gcc -O1 -fomit-frame-pointer -o irred irred.c irredasm.o
   and running on 300 MHz IBM PC (Portland) under Linux.

   The times do not include sieving (usually less than 10% of total).
   
   e.g. (r = 9689, all s < r/2)	   takes 85 sec	    or 0.091 r^3 nsec
	(r = 19937, all s < r/2)   takes 627 sec    or 0.078 r^3 nsec
	(r = 44497, all s < r/2)   takes 6627 sec   or 0.075 r^3 nsec
	
   	(r = 19937,   s = 9842)    takes 0.56 sec   or 1.41*r^2 nsec
        (r = 44497,   s = 21034)   takes 3.10 sec   or 1.57*r^2 nsec
        (r = 110503,  s = 53719)   takes 26.4 sec   or 2.16*r^2 nsec
        (r = 132049,  s = 52549)   takes 37.8 sec   or 2.17*r^2 nsec
	(r = 756839,  s = 279695)  takes 1430 sec   or 2.50*r^2 nsec
				   (overall 1625 sec)
        (r = 859433,  s = 288477)  takes 1709 sec   or 2.31*r^2 nsec (MMX2)
        			   (overall 1845 sec)

	[Following for older, obsolete version]
        (r = 3021377, s = 1234562) takes 18.3 hours or 7.23*r^2 nsec
   	 [sieving to n = 25 time 3581 sec, overall time 19.3 hours]

        (r = 6972593, s = 123444)  takes 112 hours  or 8.28*r^2 nsec (est.)

   Constants c in time c*r^2 nsec are as follows:

   300 Mhz IBM PC (Portland), CPU id 80f9ff (hex)
       CPU info: GenuineIntel
            634 0   80f9ff 0
        3020101 0  c040843 0
         
   				2.57 (r = 19937, s = 9842) 
   				        gcc -O1 -fomit-frame-pointer
   					(same for TABLEZ or not)
				2.24 (r = 19937) gcc -O1 -fomit... -funroll...
					using gcc 2.95.2 (Z's irredgz)
				1.41 (r = 19937) using ASM
			        1.57 (r = 44497) using MMX, MMX2
			        2.16 (r = 110503, s = 53719)  MMX
			        2.00 (r = 132049, s = 7000)   MMX (type 1)
			        2.17 (r = 132049, s = 52549)  MMX (type 0)

			        3.32 (r = 756839) gcc -O1 -fomit-frame-pointer
				2.50 (r = 756839, s = 279695) MMX  (type 2)
				2.59 (r = 756839, s = 279695) MMX2 (type 2) 
				2.55 (r = 756839, s = 279695) MMX2 (type 2) 
					(new XOR order)
				3.40 (r = 859433, s = 288477)
					gcc 2.7.2.3, TABLE
				2.39 (r = 859433) MMX (type 2)
					(overall 1904 sec, reducem, even delta)
				2.31 (r = 859433) MMX, MMX2
					(overall 1845 sec, reduce2, odd delta)
				2.36 (r = 859433) MMX, MMX2, prefetch 
					(prefetch has no effect, small penalty)	
				7.23 (r = 3021377, s = 1234562), TABLE
				7.40 (r = 3021377, s = 1234562), TABLEZ
				6.77 (r = 3021377, s = 1510611), MMXR (Larvala)
				6.31 (r = 3021377, s = 1510611), MMX (type 1)

   booth1-35 using NOT UNROLL and gcc 2.95.2 with
   gcc -O1 -fomit-frame-pointer -funroll-loops:

   500 Mhz IBM PC Pentium 3 Xeon (booth1-35), CPU id 383fbff (hex),
       512 KB L1 cache (full speed):
   
         CPU info: GenuineIntel
              673 0  383fbff 0
          3020101 0  c040843 0
        
   Info from Craig: Dell P III, 500 MMP with 256MB of memory and 8GB disk     
   		    installed 26 August 1999.
   		    
   (can do prefetching on booth1-35)	
   	
   				1.46 (r = 19937) TABLE (not TABLEZ), UNROLL
   				1.36 (r = 19937) TABLEZ (and TABLE), UNROLL
   				1.33 (r = 19937) TABLEZ, not UNROLL
   				1.16 (r = 19937) MMXR (Larvala)
   				1.23 (r = 19937) MMXR, MMXS
   				1.11 (r = 19937) MMXR, MMXS (Larvala 30/6)
	   			0.81 (r = 19937) MMX, MMX2
				0.95 (r = 23209, s = 6619) MMX2 type 2
				0.91 (r = 23209) MMX, MMX2
				0.89 (r = 23209) MMX2, preloading in reduce2
				1.06 (r = 23209) MMX2, 3 x pretcht0 in reduce2n
   				1.34 (r = 44497) TABLEZ, not UNROLL
   				1.09 (r = 44497) MMXR
   				1.31 (r = 44497) MMXS
   				1.16 (r = 44497) MMXR, MMXS
   				1.06 (r = 44497) MMXR, MMXS (Larvala 30/6)
   				0.94 (r = 44497) MMX (aligned)
   				1.94 (r = 110503) TABLEZ, not UNROLL
   				1.52 (r = 110503) MMXR (Larvala)
   				1.71 (r = 110503) MMXR (Brent)
   				1.34 (r = 110503) MMXR, MMXS (Larvala 30/6)
   				1.46 (r = 110503, s = 53719) MMX (aligned)
   				1.34 (r = 110503, s = 25230) MMX (aligned)
   				1.47 (r = 110503, s = 25230) MMX2
				1.29 (r = 110503, s = 25230) MMX2, preload
				1.25 (r = 110503, s = 25230) MMX2
					3 x prefetcht0 in reduce2n
				1.20 (r = 132049, s = 7000) MMX, MMX2 (type 1)
					with prefetching
				1.23 (r = 132049), MMX, MMX2, no prefetching	
			 	2.06 (r = 756839, s = 215747) TABLE, UNROLL
			 	2.11 (r = 756839, s = 215747) TABLEZ, UNROLL
				1.47 (r = 756839, s = 215747) ASM (type 2)
				1.41 (r = 756839, s = 267428) ASM (type 1)
				1.41 (r = 756839, s = 279695) ASM (type 1)
			 	1.81 (r = 859433, s = 288477) TABLEZ
			 	1.83 (r = 859433) TABLE, UNROLL
			 	1.96 (r = 859433) TABLEZ, UNROLL
				1.38 (r = 859433, s = 170340) ASM (17 min)
				1.42 (r = 859433, s = 288477) ASM (< 18 min)
				3.43 (r = 3021377, s = 1234562) TABLEZ
				2.69 (r = 3021377, s = 1234562) MMX2 (type 1)
				2.37 (r = 3021377, s = 1234562) MMX2, prefetch	
					(6.67 hours overall)
				2.38 (r = 3021377, s = 1234562) MMX2, prefetch,
					fast reducer (6.42 hours overall)
				2.57 (r = 3021377, s = 1510611) MMX2 (type 1)
				2.48 (r = 3021377, s = 1510611) MMX2, prefetch	
					(6.33 hours overall)
				2.39 (r = 3021377, s = 390000) MMX2, prefetch but
					not in squarem (6.30 hours overall)
				2.24 (r = 3021377, s = 390000) MMX2, prefetch
					including one in squarem
				2.24 (r = 3021377, s = 390030) MMX2, 
					(5.92 hours overall)
					prefetcht0 four ahead (i.e. one cache 
					line ahead)
				2.25 (r = 3021377, s = 390030) MMX2,
					prefetcht0 eight ahead (2 cache lines)
				2.65 (r = 3021377, s = 390030) MMX2,
					no prefetching
				2.52 (r = 3021377, s = 390209) MMX2 - this is a
					case r-s = 0 (mod 32) so r2 = 0 and
					assembler not used (though could be ?).
				2.23 (r = 3021377, s = 361604, primitive) MMX2
					20330 sec, 21090 sec overall
												
   bsp0..7 PC cluster with egcs-2.91.66, use	
   gcc -O2 -fomit-frame-pointer -funroll-loops
   bsp0 400MHz PII IBM PC, 320 MB, CPU id 183f9ff (hex). 	
   
   CPU info: GenuineIntel
        651 0  183f9ff 0
    3020101 0  c040843 0
   				1.64 (r = 19937, s = 9842) TABLEZ
   				1.11 (r = 19937) ASM
				2.15 (r = 110503, 25230) TABLEZ 
				1.77 (r = 110503) ASM
				1.92 (r = 859433, s = 288477) MMX, MMX2
				1.97 (r = 859433) MMX2, preload in reduce2
				2.88 (r = 3021377, s = 120020) ASM	
									 
   Pentium PIII-366 laptop under Linux (Zimmermann) with gcc 2.95.2
    				2.05 (r = 110503)
    				2.19 (r = 756839) TABLE
    				2.21 (r = 756839) TABLEZ

   Sun gauss  300 Mhz Sparc	2.74 (r = 19937) with gcc 2.95.2, -O6
   				4.15 (r = 3021377, s = 1510653)
				4.19 (r = 3021377, s = 361604, primitive)
   Sun janus  250 Mhz Sparc	3.30 (r = 19937) with gcc 2.95.1, -O6
                                3.89 (r = 3021377, s = 390000)
   Sun discus 175 Mhz Sparc	5.11 (r = 19937) with gcc 2.95.1, -O6

   SGI R10000 (32-bit version)	2.99 (TABLE,UNROLL), 3.40 

64-bit version (W64 = TRUE):
   
   Times (excluding sieving) on SGI R10000 compiled with cc -64 -Ofast

   	1.81*r^2 for r = 19937,  s = 9842
   	1.82     for r = 110503, s = 53719
	2.24     for r = 756839, s = 279695
	2.29     for r = 859433, s = 288477
	2.23     for r = 3021377, s = 1234562
                [sieving time to n = 25 is 4142 sec, overall 24524 sec]
	2.23     for r = 3021377, s = 361604 (primitive)
		[sieving time to n = 25 is 806 sec, overall 21166 sec] 			
   	
   Times for version 1.30 on 500Mhz DEC Alpha EV6 DEC
   compiled with gcc -O4 -mcpu=ev6 -Wa,-arch -Wa,ev6 -funroll-loops
   (from Paul Zimmermann)

	0.59*r^2 for r = 110503, s = 53719  (7.63 = 7.20 + 0.43 sec)
   	0.92     for r = 756839, s = 279695 (554 = 528 + 26 seconds)
        0.92	 for r = 859433, s = 288477 (700 = 681 + 19 seconds)
        0.88	 for r = 3021377, s = 390000 (8604 = 8027 + 577 sec)

References:

   J. W. Heringa, H. W. J. Bl\"ote and A. Compagner, International Journal
    of Modern Physics C 3 (1992), 561-564.

   D. E. Knuth, The Art of Computer Programming, Vol. 2, third ed., Sec. 3.2.2.

   T. Kumada, H. Leeb, Y. Kurita and M. Matsumoto,
    Mathematics of Computation 69 (2000), 811-814;

   N. Zierler, Information and Control 15 (1969), 67.

Comments on different versions:

   Version 1.35 has
   
   1) Better detection of illegal parameters (if IBMPC AND ASM)

   Version 1.34 has
   
   1) More informative output re CPU type (if IBM PC).

   Version 1.33 has
   
   1) Faster code in case that r-s is divisible by 32 and IBM PC assembler
      is available (previously this case did not call assembler).

   2) ASM flag reinstated to replace MMX and MMX2 flags.

   Version 1.32 has
   
   1) More informative output - times for last sieve iteration
      and fraction of time spent sieving.

   2) Bug in versions 1.30 & 1.31 (which gave wrong "n" for log on stdout
      if printed in phase 2) fixed.
      
   3) Storage allocation bug (which caused segmentation fault if only one
      command-line argument on some machines) fixed.
        
   Version 1.31 has
   
   1) Unrolled option in reducer (may help on Sun etc).

   Version 1.30 has 
   
   1) Faster relprime calling reducer for inner loop
      (with assembler version for IBM PC).
   2) Introduction of NEXTRA and abolition of RATIO.
   3) Storage requirement reduced to 3r/8 + O(1) bytes.

   Version 1.25 has
   
   1) More CPU information in case of IBM PC.
   2) Two assembler versions - 
      one (irrednas.s) uses prefetching (works on PIII and above)
      the other (irredasm.s) without prefetching.

   Version 1.24 has
   
   1) MMX2 option to allow reduce2.s in case q1 and q1-q2 both odd.

   Version 1.23 has
   
   1) Alignment of most 8-byte accesses on 8-byte boundaries.
   2) Abolition of PORTLAND, ASM, OLDGCC, OLDR options.
   3) Merging of MMXR and MMXS options to MMX.

   Version 1.22 has
   
   1) Larvala's new MMX routines (renamed squarem and reducem)

   Version 1.21 has
   
   1) Alternative version of MMX (reduce2.s) modified RPB  

   Version 1.20 has
   
   1) MMXR and MMXS options for use of Multi Media extension (mmx) 
      instructions on Pentium etc (written by Samuli Larvala).

   Version 1.12 has
   
   1) Slightly faster reducea (though still some room for improvement ?)
      Set OLDR to use the old version of reducea (may be faster on machines
      other than IBM PC).
      
   2) Comments and output improved (2^p is never prime for p > 1 !).   

   Version 1.11 has
       
   1) ASM flag added and square.s implemented (IBM PC version only)
   2) comments improved 

   Compared to the previous versions 1.0 and earlier (unnumbered),
   version 1.1 has

   1) a fix to avoid timer overflow for large r.
   2) memory requirements for sieving reduced by about 46%.
   3) addition of a "skip file" as optional 4th command-line argument.
   4) various small enhancements and improvements in the comments.

Compilation flags:

   Set following flags to optimise performance depending on machine type
   and compiler. The best settings for TABLE, TABLEZ, UNROLL
   have been determined by timing experiments and could change with
   different compilers or variations in hardware. 
   
   On an IBM PC which can execute MMX instructions it is best 
   to set ASM = TRUE. 
   
   On an IBM PC P-III and above which can implement prefetching it
   is best to use assembler routines irrednas.s/irrednas.o
   For processors earlier than P-III, use irredasm.s/irredasm.o .

*/


#define ASM	IBMPC			   /* Use external assembler versions 
					      of square and reducea.
					      At present these are only 
					      available for the IBM PC and
					      assume a Pentium or above as
					      they use MMX instructions. */

#define W64   	(ALPHA OR SGI)		   /* For 64-bit version. Faster than
					      32-bit version if available. */

#define TABLE ((IBMPC OR SGI) AND NOT W64) /* To use table lookup for square.
					      Not implemented for 64-bit
					      versions (as too slow then). */

#define TABLEZ IBMPC 			    /* To use little-endian 
					      version of table lookup 
					      (if TABLE also set; irrelevant
					      if IBMPC AND ASM) */

#define PRELOAD SPARC			   /* To use preload in reducea */

#define UNROLL (ALPHA OR SGI OR SPARC)	   /* To use unrolled square. */

#if IBMPC
 #define REGISTER			   /* Omit register declaration on PC */
#else
 #define REGISTER register		   /* Include it on other machines */	
#endif

/* Some constants */

#define TIMEOUT 60		/* Seconds to timeout when opening files */
#define CPUtol 0.01		/* Times less than CPUtol sec are negligible */

#define EQ ==
#define NE !=
#define GE >=
#define GT >
#define LE <=
#define LT <
#define ERROR 1
#define OK 0

/* Type definitions */

typedef char BOOLEAN;
typedef unsigned char  UCHAR;	/* Assumed to be 8 bits */
typedef unsigned short USHORT;  /* Should be 16 bits for TABLEZ option */
typedef unsigned int   UINT;	/* Assumed to be (at least) 32 bits */  	
typedef unsigned long  ULONG;	/* Should be 32 or 64 bits */

#define WLEN  (8*sizeof(ULONG))	/* Bits in a long word, should be 32 or 64 */
#define WLENM (WLEN-1)		/* Ditto less 1, i.e. 31 or 63 */
#define WD    (W64 + 5)		/* 5 or 6, so WLEN = 2^WD */
  
#if (ASM AND IBMPC)
  #define WLEN2 (2*WLEN)	/* Effective wordlength 64 for MMX */
#else
  #define WLEN2 WLEN
#endif      
  
#define NMAX  32		/* Sieve over x^kn - 1 for kn = 2^n-1,
				   n LE NMAX (if r large enough). */
	
#define	NEXTRA 4		/* Do NEXTRA GCD computations for 
				   2^(degree) GE r. Empirically we find
				   that 4 or 5 is about optimal (depending on
				   the machine), because the density of nonzeros 
				   in the polynomials and the time for GCDs 
				   increases as the degree increases. 
				   Use NEXTRA = 4 to avoid differences
				   in output on different machines. */
				   
#define MC    100		/* Max input line length */

/* Struct for the skip list (a singly-linked list of pairs) */

struct skip {
  int low, high;
  struct skip *next;
  };

/* Global variables */

int r, s, q1, r1, r1c, q2, r2, r2c;	/* Depend only on r and s */

double CPUtime;				/* For timer */
clock_t cstart;

int space = 0;				/* For malloc count */

/* Routines start here */

double clockd(starta, first)
clock_t *starta;
BOOLEAN first;

/* If first = TRUE, initialises *starta and returns zero.
   If first = FALSE, returns time in sec since last call and updates starta. 
   Note that this routine should be called at least each 4290 sec on
   IBM PC (under Linux) since CLOCKS_PER_SEC = 1000000 and 32-bit integer
   overflow may occur if called infrequently */
  
  {
  clock_t stop;
  ULONG diff;
  stop = (clock_t) clock();
  if (first) *starta = stop;
  diff = (ULONG)(stop - *starta);	/* May wrap around in 2^32/10^6 sec */
  *starta = stop;
  return ((double)diff/(double)CLOCKS_PER_SEC);
  }


char *mymalloc(size)
int size;

/* Same as malloc but gives error exit if can not allocate space */

  {
  char *ptr;
  space += size;
  ptr = (char *)malloc(size); 
  if (ptr EQ NULL) {
    printf("malloc unable to allocate %d bytes\n", space);
    exit(ERROR);
    }
  return(ptr);
  }  
    
BOOLEAN comparex(a)

/* Returns TRUE if poly a of degree r-1 is x */

ULONG *a;
  {
  int j;
  if (a[0] NE 2L) return(FALSE);
  for (j = q1; j > 0; j--) {
    if (a[j] NE 0) return(FALSE);
    }
  return(TRUE);
  }
  
#if (IBMPC AND ASM)

extern void reducer(ULONG* a, ULONG* b, int kt, int rs, ULONG* prev);

#elif UNROLL

void reducer(a, b, kt, shift, prev)	

/* Unrolled version.
   Simplified version of reducea, called by relprime.
   Assumes 0 < shift < WLEN. */

ULONG *a, *b;
int kt, shift;
ULONG *prev;		/* Previous -> last value of new */

  {
  REGISTER int j, shiftc;
  REGISTER ULONG old, new;
  new = *prev;
  shiftc = WLEN - shift;
  for (j = kt; (j GE 0) AND ((j & 1) EQ 0); j--) { /* One iteration if
  						      kt even */
    old = new;
    new = a[j];
    b[j] ^= (new >> shift) | (old << shiftc);
    }
  for (; j GE 0; j -= 2) {
    old = a[j];					/* old <-> new here ! */
    b[j] ^= (old >> shift) | (new << shiftc);
    new = a[j-1];				
    b[j-1] ^= (new >> shift) | (old << shiftc);
    }
  *prev = new;
  return;  
  }  
  
#else  

void reducer(a, b, kt, shift, prev)	

/* Simplified version of reducea, called by relprime.
   Assumes 0 < shift < WLEN. */

ULONG *a, *b;
int kt, shift;
ULONG *prev;		/* Previous -> last value of new */

  {
  REGISTER int j;
  REGISTER ULONG old, new;
  for (new = *prev, j = kt; j GE 0; j--) {
    old = new;
    new = a[j];
    b[j] ^= (new >> shift) | (old << (WLEN-shift));
    }
  *prev = new;
  return;  
  }  
  
#endif

BOOLEAN relprime (p, np, q, nq)

/* Returns TRUE if polynomials p and q (of degrees np and nq)
   are relatively prime (mod 2). Destroys p and q in the process.
   Returns FALSE if np < 0 or nq < 0.
   Time is O(np.nq). */
   
ULONG *p, *q;
int np, nq;
    
  {
  int j, k, del, delq, delr, delrc, high;
  int wkt = 0;
  REGISTER ULONG new;
  ULONG *t;
  ULONG prev;

  if ((np LT 0) OR (nq LT 0)) return(FALSE);
  
  while (TRUE) {
    if ((wkt -= nq) < 0) {
      CPUtime += clockd(&cstart, FALSE);
      					/* Avoid timer overflow by calling */
      wkt = 0x40000000; 		/* clockd occasionally */
      } 
    if (np LT nq) { 			/* p <-> q etc */
      k = np; np = nq; nq = k;
      t = p;  p = q; q = t;
      }  
    del = np - nq;
    delq = del >> WD;			/* delq = del/WLEN */			
    delr = del & WLENM;			/* delr = del%WLEN */			
    if (delr EQ 0) {			/* Simple case, word aligned */
      for (j = (nq >> WD); j GE 0; j--) 
        p[j+delq] ^= q[j];  
      }
    else {  				/* Need bit-alignment here */
      delrc = WLEN - delr;		/* in 1..(WLEN-1) since delr NE 0 */
      high = (np >> WD) - delq;
      prev = q[high];
      reducer(q, p+delq+1, high-1, delrc, &prev);
      p[delq] ^= (prev<<delr);
      }    
    for (j = np; j GE 0; j--) { /* Find new deg(p) */
      new = p[j >> WD];
      if (new EQ 0L) {		/* Speed up search over zero words */
        np -= j & WLENM;
        j-= j & WLENM;		
	}
      else
        if ((new >> (j & WLENM)) &1L) break;
      np--;
      }
    for (j = nq; j GE 0; j--) {	/* and deg(q) (not usually necessary) */
      if ((q[j >> WD] >> (j & WLENM)) &1L) break;
      nq--;
      }
    if (np LT 0) return (nq EQ 0);
    }
  }    

void setupx(a)

/* Sets up a = x (polynomial of degree r-1, all mod 2) */

ULONG *a;
  {
  int j;
  for (j = q1; j GT 0; j--) 
    a[j] = 0L;
  a[0] = 2L;			/* 0...010 represents x */
  }  

#if (ASM AND IBMPC)

/* Here use assembler version squarem of square which uses the MMX
   instructions (available on most Pentiums and above, but not 486 machines).
   This routine was written by Samuli Larvala. */
   
extern void squarem(ULONG* a, int q1);

extern void cpuinfo(int flag, int* result); 
				/* Returns info from cpuid instruction */
extern BOOLEAN checkmmx();	/* Returns TRUE if MMX instruction available */
extern BOOLEAN prefetching();	/* Returns TRUE if prefetching versions of
				   assembler routines (irrednas.s) used */

#elif (TABLE AND TABLEZ)

void squaret (a, tabw, q1)

ULONG *a; 
USHORT *tabw;
int q1;
   
/* 8-bit -> 16-bit table lookup as suggested by Paul Zimmermann.
   Faster than my original 16-bit -> 32-bit table lookup (on IBM PC).
   Assumes little-endian byte order. */
   
  {
  UCHAR *aa = (UCHAR *) a;
  USHORT *b = (USHORT *) a;
  long j;

  aa += 4*(q1+1);		  /* Could set some constants up as */
  b  += 4*(q1+1);		  /* globals but would not save much */
#if UNROLL
  for (j = 2*q1+1; j GE 0; j--) { /* This helps with gcc 2.7.2.3 but */
    *--b = tabw[*--aa];		  /* not with gcc 2.95.2 */
    *--b = tabw[*--aa];
    }
#else
  for (j = 4*q1+3; j GE 0; j--) { 
    *--b = tabw[*--aa];		
    }
#endif
  }

#elif TABLE

void squaret (a, tabw, q1)
ULONG *a, *tabw;
int q1;
   
/* 8-bit -> 16-bit table lookup for 32-bit machines
   (works on both big-endian and little-endian machines).
   May be slower/faster than Zimmermann's little-endian version (above)
   depending on the compiler, but not much difference for large r,
   when performance is limited by memory bandwidth. */
   
  {
  REGISTER ULONG temp;
  REGISTER long j;
  for (j = q1; j GE 0; j--) { 
    temp = a[j];
    a[2*j]   = tabw[(UCHAR)temp] | 
    	      (tabw[(UCHAR)(temp >> 8)] << 16);
    a[2*j+1] = tabw[(UCHAR)(temp >> 16)] |
              (tabw[(UCHAR)(temp >> 24)] << 16);
    }
  }

#elif (W64 AND (NOT UNROLL))

void square(a)			/* 64-bit version */

/* Squares polynomial a of degree r-1 (mod 2), result -> a,
   without table lookup, assuming 64-bit words.
   (1.load + 2.store).q1 memory ops. and about 30.q1 REGISTER ops.
   NB: Because of an optimisation in the loop, it should be valid
   to read a[-1]. */

ULONG *a;
  {
  REGISTER int j;
  REGISTER ULONG t, u, next;
  REGISTER ULONG c0, c1, c2, c3, c4, c5;
  
  c0 = 0x00000000FFFFFFFFL;		/* Some 64-bit constants */
  c1 = 0x0000FFFF0000FFFFL;
  c2 = 0x00FF00FF00FF00FFL;
  c3 = 0x0F0F0F0F0F0F0F0FL;
  c4 = 0x3333333333333333L;
  c5 = 0x2222222222222222L;
    
  next = a[q1];
  
  for (j = q1; j GE 0; j--) {

    u = next>>32;			/* High order 32 bits of a[j] */
    t = next&c0;			/* Low order 32 bits */

    u = (u | u<<16) & c1;		 /* Operations on t and u can be */
    t = (t | t<<16) & c1;		 /* done in parallel */
    
    next = a[j-1];			 /* Do load early to overcome memory
    					    latency. May access a[-1] ! */
    u = (u | u<<8)  & c2;	
    t = (t | t<<8)  & c2;	
    u = (u | u<<4)  & c3;
    t = (t | t<<4)  & c3;
    u = (u | u<<2)  & c4;
    t = (t | t<<2)  & c4;
    
    a[2*j+1] = u + (u & c5);
    a[2*j]   = t + (t & c5);
    }
  }

#elif (W64 AND UNROLL)

void square(a)				/* 64-bit unrolled version */

/* Unrolled version of square.
   Squares polynomial a of degree r-1 (mod 2), result -> a,
   without table lookup, assuming 64-bit words.
   (1.load + 2.store).q1 memory ops. and about 30.q1 REGISTER ops.

   May be faster than "un-unrolled" version because more than two 
   REGISTER ops can be performed concurrently on machines such as
   SGI R10000 and DEC alpha.

   Because of a loop optimisation, it should be valid to access a[-2].
   
   IBM PC assembler version of this using MMX instructions is squarem.s
   (can be used without setting W64). 
*/

ULONG *a;
  {
  REGISTER int j, low;
  REGISTER ULONG t, u, v, w, next1, next2;
  REGISTER ULONG c0, c1, c2, c3, c4, c5;
  
  c0 = 0x00000000FFFFFFFFL;		/* Some 64-bit constants */
  c1 = 0x0000FFFF0000FFFFL;
  c2 = 0x00FF00FF00FF00FFL;
  c3 = 0x0F0F0F0F0F0F0F0FL;
  c4 = 0x3333333333333333L;
  c5 = 0x2222222222222222L;
  
  low = 2*((q1+1)/2) - 1;		/* Assume q1 GT 0 */
    
  if (q1 GT low) {			/* Do an iteration of loop if q1 even */

    t = a[q1];
    u = t>>32;				/* High order 32 bits of a[q1] */
    t &= c0;				/* Low order 32 bits */

    u = (u | u<<16) & c1;		 /* Operations on t and u can be */
    t = (t | t<<16) & c1;		 /* done in parallel */
    u = (u | u<<8)  & c2;	
    t = (t | t<<8)  & c2;	
    u = (u | u<<4)  & c3;
    t = (t | t<<4)  & c3;
    u = (u | u<<2)  & c4;
    t = (t | t<<2)  & c4;
    a[2*q1+1] = u + (u & c5);
    a[2*q1]   = t + (t & c5);
    }

  next1 = a[low];
  next2 = a[low-1];
  
  for (j = low; j GE 0; j -= 2) {	/* Here low and j are odd */
				
    u = next1>>32;			/* High order 32 bits of a[j] */
    w = next2>>32;			/* Ditto of a[j-1] */
    t = next1 & c0;			/* Low order 32 bits of a[j]*/
    v = next2 & c0;			/* Ditto of a[j-1] */

    u = (u | u<<16) & c1;		 /* Operations on t,u,v,w   */
    t = (t | t<<16) & c1;		 /* can be done in parallel */
    w = (w | w<<16) & c1;
    v = (v | v<<16) & c1;

    next1 = a[j-2];			 /* Perform loads in advance 
    					    to overcome latency */
    u = (u | u<<8) & c2;	
    t = (t | t<<8) & c2;	
    w = (w | w<<8) & c2;	
    v = (v | v<<8) & c2;	

    next2 = a[j-3];			 /* Ditto (may access a[-2]) */ 
    
    u = (u | u<<4) & c3;
    t = (t | t<<4) & c3;
    w = (w | w<<4) & c3;
    v = (v | v<<4) & c3;

    u = (u | u<<2) & c4;
    t = (t | t<<2) & c4;
    w = (w | w<<2) & c4;
    v = (v | v<<2) & c4;

    a[2*j+1] = u + (u & c5);
    a[2*j]   = t + (t & c5);
    a[2*j-1] = w + (w & c5);
    a[2*j-2] = v + (v & c5);
    }
  }
  
#elif (NOT UNROLL)

void square(a)				/* 32-bit version */

/* Squares polynomial a of degree r-1 (mod 2), result -> a,
   without table lookup, assuming 32-bit words.
   (1.load + 2.store).q1 memory ops. and about 24.q1 REGISTER ops. */

ULONG *a;
  {
  REGISTER int j;
  REGISTER ULONG t, u, next;
  REGISTER ULONG c1, c2, c3, c4, c5;
  
  c1 = 0x0000FFFFL;		/* Some 32-bit constants */
  c2 = 0x00FF00FFL;
  c3 = 0x0F0F0F0FL;
  c4 = 0x33333333L;
  c5 = 0x22222222L;
    
  next = a[q1];
    
  for (j = q1; j GE 0; j--) {

    u = next>>16;				/* High order 16 bits of a[j] */
    t = next&c1;				/* Low order 16 bits */

    u = (u | u<<8)  & c2;	
    t = (t | t<<8)  & c2;	
    
    next = a[j-1];				/* Load in advance to overcome
    						   latency. May access a[-1] */
    u = (u | u<<4)  & c3;
    t = (t | t<<4)  & c3;
    u = (u | u<<2)  & c4;
    t = (t | t<<2)  & c4;
    a[2*j+1] = u + (u & c5);
    a[2*j]   = t + (t & c5);
    }
  }
  
#else  

void square(a)					/* 32-bit unrolled version */

/* Unrolled version of square.
   Squares polynomial a of degree r-1 (mod 2), result -> a,
   without table lookup, assuming 64-bit words.
   (1.load + 2.store).q1 memory ops. and about 24.q1 REGISTER ops.
   
   May be faster than "un-unrolled" version if more than two REGISTER ops can
   be performed concurrently. Faster than table lookup on word-oriented
   machines (e.g. SPARC) but not on IBM PC. 
   
   Because of a loop optimisation, it should be valid to access a[-2]. */

ULONG *a;
  {
  REGISTER int j, low;
  REGISTER ULONG t, u, v, w, next1, next2;
  REGISTER ULONG c1, c2, c3, c4, c5;
  
  c1 = 0x0000FFFFL;			/* Some 32-bit constants */
  c2 = 0x00FF00FFL;
  c3 = 0x0F0F0F0FL;
  c4 = 0x33333333L;
  c5 = 0x22222222L;
  
  low = 2*((q1+1)/2) - 1;		/* Assume q1 GT 0 */
    
  if (q1 GT low) {			/* Do an iteration of loop if q1 even */

    t = a[q1];
    u = t>>16;				/* High order 16 bits of a[q1] */
    t &= c1;				/* Low order 16 bits */

    u = (u | u<<8)  & c2;	
    t = (t | t<<8)  & c2;	
    u = (u | u<<4)  & c3;
    t = (t | t<<4)  & c3;
    u = (u | u<<2)  & c4;
    t = (t | t<<2)  & c4;
    a[2*q1+1] = u + (u & c5);
    a[2*q1]   = t + (t & c5);
    }

  next1 = a[low];
  next2 = a[low-1];
  
  for (j = low; j GE 0; j -= 2) {	/* Here low and j are odd */
				
    u = next1>>16;			/* High order 16 bits of a[j] */
    w = next2>>16;			/* Ditto of a[j-1] */
    t = next1 & c1;			/* Low order 16 bits of a[j]*/
    v = next2 & c1;			/* Ditto of a[j-1] */

    u = (u | u<<8) & c2;	
    t = (t | t<<8) & c2;	

    next1 = a[j-2];			 /* Perform loads in advance 
    					    to overcome latency */
    w = (w | w<<8) & c2;	
    v = (v | v<<8) & c2;	

    next2 = a[j-3];			 /* Ditto (may access a[-2]) */ 
    
    u = (u | u<<4) & c3;
    t = (t | t<<4) & c3;
    w = (w | w<<4) & c3;
    v = (v | v<<4) & c3;

    u = (u | u<<2) & c4;
    t = (t | t<<2) & c4;
    w = (w | w<<2) & c4;
    v = (v | v<<2) & c4;

    a[2*j+1] = u + (u & c5);
    a[2*j]   = t + (t & c5);
    a[2*j-1] = w + (w & c5);
    a[2*j-2] = v + (v & c5);
    }
  }

#endif

#if (IBMPC AND ASM)

/* Here use assembler version reducem of reducea which uses the MMX
   instructions (available on most Pentiums and above, but not 486 machines).
   This routine was written by Samuli Larvala. */
   
extern ULONG reducem(ULONG* a, ULONG* b, ULONG* c, int r1, int r2, int kt);

/* reduce2 is a slight variation on reducem, faster in some cases */

extern ULONG reduce2(ULONG* a, ULONG* b, ULONG* c, int r1, int r2, int kt);

#endif

ULONG reducea(a, b, c, kt, prev)	

/* Fast inner loop for reduce. Assumes r1 nonzero.
   For IBM PC assembler version, see reducem.s */

ULONG *a, *b, *c;
int kt;
ULONG prev;		/* Previous value of new */

  {
  REGISTER int j;
  REGISTER ULONG old, new;
#if PRELOAD
  REGISTER ULONG bj, cj;
#endif  
  if (r2 EQ 0) {	/* Occurs with probability 1/WLEN */
    for (new = prev, j = kt; j GE 0; j--) {
      old = new;
      new = a[j];
      b[j] ^= (new>>r1) | (old<<r1c);
      c[j] ^= new;	/* Need special case as old << r2c must be zero */
      }
    }
  else {
    for (new = prev, j = kt; j GE 0; j--) {
      old = new;
      new = a[j];
#if PRELOAD
      bj = b[j];
      cj = c[j];      
      b[j] = bj^((new>>r1) | (old<<r1c));
      c[j] = cj^((new>>r2) | (old<<r2c));
#else
      b[j] ^= (new>>r1) | (old<<r1c);
      c[j] ^= (new>>r2) | (old<<r2c);
#endif
      }
    }
  return(new);  
  }  

void reduce(a)

/* Reduces polynomial a of degree < 2r-1 given x^r = x^s + 1 (mod 2),
   result -> a.
   Assumes 0 < s < r AND r-s GE WLEN. 
   (3.load + 2.store).q1 memory ops. */

ULONG *a;

  {
  REGISTER ULONG old, new;
  REGISTER int j;
  new = 0;

  /* Have to check cases r1 or r2 EQ 0,
     i.e. r1c or r2c EQ WLEN, because << WLEN is NOP on some machines ! */
  
  if (r1 EQ 0) {	/* This never occurs if r odd */
    for  (j = 2*q1+1; j > q1; j--) {
      old = new;
      new = a[j];
      a[j-q1] ^= new;
      a[j-q2] ^= (r2 EQ 0) ? new : (new>>r2) | (old<<r2c);
      }
    }

  else if (q1 EQ q2) {			/* Fast loop for q1 EQ q2 */

    if (r2 EQ 0) { 			/* Need special case here */
      for  (j = 2*q1+1; j > q1; j--) {
        old = new;
        new = a[j];
        a[j-q1] ^= ((new>>r1) | (old<<r1c))^new;
	}
      }
    else {  
      for  (j = 2*q1+1; j > q1; j--) {
        old = new;
        new = a[j];
        a[j-q1] ^= ((new>>r1) | (old<<r1c))^((new>>r2) | (old<<r2c));
	}
      }
    }
      
  else if ((q1-q2) LE 1) {		/* reducem doesn't work here ! 
    					   e.g. try r = 521, s = 32 */
    new = reducea(a+1+q1, a+1, a+1+q1-q2, q1, 0L);
    }
    
  else {		/* Usual case, call reducem/reduce2/reducea */

#if (IBMPC AND ASM)	/* Assembler version using MMX instructions */

    if (((q1 & 1) + ((q1-q2) & 1)) LE 1)  
      {
      /* This is to make most array accesses on 8-byte boundaries.
         Here at most one of q1, q1-q2 is odd, and seems best to 
         use reducem with even displacement. */
      
      new = reducem(a+2+q1, a+2, a+2+q1-q2, r1, r2, q1-1);
      
      /* Need one more iteration - could inline it */
      
      new = reducea(a+1+q1, a+1, a+1+q1-q2, 0, new);
      }
    else {	
    
   /* Here both q1 and q1-q2 are odd.
      Tests with r = 859433, s = 288477 indicate that it is slightly
      faster to use reduce2 than reducem/reducea as above.
   */
      new = reduce2(a+1+q1, a+1, a+1+q1-q2, r1, r2, q1);
      }
#else				/* C version */

    new = reducea(a+1+q1, a+1, a+1+q1-q2, q1, 0L);

#endif
    }

  /* Cleanup - last one/two iterations are special. */

  old = new;
  new = (a[q1] >> r1) << r1;	/* Clear degree < r terms */
  a[0]     ^= (r1 EQ 0) ? new : (new >> r1) | (old << r1c);
  a[q1-q2] ^= (r2 EQ 0) ? new : (new >> r2) | (old << r2c);
  if ((q1 > q2) AND (r2 NE 0)) a[q1-q2-1] ^= (new<<r2c);
  a[q1] &= ~(~0L << r1); 	/* Clear any degree GE r terms */
  }
  
FILE *myfopen(fname, flag)

/* Attempts to open file fname. If not successful at first, keeps trying
   for TIMEOUT seconds, then error exits.
*/
   
char *fname, *flag;

  {
  FILE *fp, *fopen();
  clock_t openstart;
  double time;
  time = clockd(&openstart, TRUE);	/* Independent timer here */
  
  for (;;) {
    fp = fopen(fname, flag);
    if (fp NE NULL) break;
    if ((time += clockd(&openstart, FALSE)) GT TIMEOUT) {
      printf("Could not open %s after trying for %d seconds\n", 
	fname, TIMEOUT);
      exit(ERROR);
      }
    }
  return(fp);
  }

BOOLEAN skips(skiplist, s)
struct skip *skiplist;
int s;

/* Returns TRUE if s is in the skip list */

  {
  struct skip *skiprec = skiplist;
  while (skiprec NE NULL) {
    if ((skiprec->low LE s) AND (s LE skiprec->high))
      return(TRUE);
    skiprec = skiprec->next;
    }  
  return(FALSE);     
  }  

int main(argc, argv)
int argc;
char *argv[];
  {
  FILE *fp;
  double CPUtime1, CPUlast, CPUtotal = 0;
#if VERBOSE
  double CPUtotal1 = 0;
#endif    
  int minutes = 0;		/* minutes before stopping */
  int j, k;
  int s1 = 0, s2 = 0;
  int n, nsv, nm, kn, pn, qn, rk, sk, sizea, sizep;
  int kt1 = 0, kt2 = 0, skt = 0;
  int temp;
  ULONG *a;			/* For polynomial of degree 2(r-1) */
  ULONG *p, *q;			/* For sieving */
  ULONG new;

  int sieved[NMAX+1];		/* Count sieve successes */
  char line[MC];		/* Line buffer */
  char log[] = " LOG";		/* Identifier for log 
  				   (if no log file specified) */
  struct skip *skiplist;	/* The head of the skip list */
  struct skip *skiprec;
  int slow, shigh;
  BOOLEAN done, found, g, swan;
  
#if (ASM AND IBMPC)
  int cpumax;
  int cpudata[5];		/* For results of cpuinfo */
#endif

#if (TABLE AND TABLEZ)
  USHORT tabw[256];		/* Lookup table for square (IBMPC) */
  UCHAR *charptr;
#elif TABLE
  ULONG tabw[256];		/* Lookup table for square (other) */
#endif

  printf("\nThis is irred version 1.35\n");	  /* Date 20000814 */

  printf("\nCopyright (C) 2000 R. P. Brent.\n");
  printf("irred comes with absolutely no warranty. This is free software,\n");
  printf("and you are welcome to redistribute it under the conditions of\n");
  printf("the GNU General Public License, version 2, June 1991.\n");
  printf("See http://www.gnu.org/copyleft/gpl.html for further details.\n\n");

  printf("Options ");
  if (IBMPC) printf("IBMPC, ");
  if (ALPHA) printf("ALPHA, ");
  if (SPARC) printf("SPARC, ");
  if (SGI)   printf("SGI, ");
  if ((IBMPC + ALPHA + SPARC + SGI) NE 1) {
    printf("\nPlease set one machine type to TRUE, ");
    printf("others to FALSE, and recompile\n");
    return(ERROR);
    }
  if (W64)   printf("W64, ");
  if (IBMPC AND ASM)
    printf("ASM, "); 
  else {
    if (TABLE AND TABLEZ) 
      printf("TABLEZ, ");
    else 
      if (TABLE) printf("TABLE, ");
    }  
  if (UNROLL) printf("UNROLL, ");
  if (PRELOAD) printf("PRELOAD, ");
  printf("NEXTRA = %d\n", NEXTRA);

#if (ASM AND IBMPC)
  if (prefetching()) 
    printf("Using prefetching versions of assembler routines\n");
  cpuinfo (0, cpudata);
  cpumax = cpudata[0];
  cpudata[4] = 0;
  printf("\nCPU info: %12s\n", (char *)(cpudata+1));
  for (j = 1; j LE cpumax; j++) {
    cpuinfo(j, cpudata);
    printf("%8x %x %8x %x\n", 
      cpudata[0], cpudata[1], cpudata[2], cpudata[3]);
    }  
    
  cpuinfo(1, cpudata);
  if (cpudata[0] EQ 0x634)
    printf("P-II similar to Portland\n");
  else if (cpudata[0] EQ 0x651)
    printf("P-II similar to bsp0\n");
  else if (cpudata[0] EQ 0x673)
    printf("P-III similar to booth1-35\n");    
#endif

  printf("\n");
      
#if W64

  if (sizeof(ULONG) NE 8) {	/* This is a 64-bit version */
    printf("sizeof(ULONG) %d, should be 8\n", (int)sizeof(ULONG));
    printf("recompile with 64-bit options\n");
    return(ERROR);
    }

#else

  if (sizeof(ULONG) NE 4) {	/* This is a 32-bit version */
    printf("sizeof(ULONG) %d, should be 4\n", (int)sizeof(ULONG));
    printf("recompile with 32-bit options\n");
    return(ERROR);
    }
#endif

  CPUtime = clockd(&cstart, TRUE); 		/* Initialise timer */
  a = NULL;
  skiplist = NULL;
  found = FALSE;
  r = 0;
  line[0] = '\0';

  for (j = NMAX; j GE 0; j--) 
    sieved[j] = 0;

#if (ASM AND IBMPC)
  if (NOT checkmmx()) {
    printf ("Sorry, your PC can not execute MMX instructions\n");
    printf ("Set ASM to FALSE and recompile\n");
    return(ERROR);
    }
#endif   

#if (TABLE AND W64)

  printf ("Sorry, TABLE is not implemented in 64-bit version\n");
  printf ("Set TABLE or W64 to FALSE and recompile\n");
  return(ERROR);

#elif (TABLE AND TABLEZ)

  if (sizeof(USHORT) NE 2) {	/* Check size, should be 16 bits */
    printf("sizeof(unsigned short) %d, should be 2\n", 
     (int)sizeof(USHORT));
    printf("Set TABLEZ to FALSE and recompile\n"); 
    return(ERROR);
    }
    
  for (j = 255; j GE 0; j--) 		/* Set up USHORT lookup table tabw */
    for (tabw[j] = 0, k = 7; k GE 0; k--)
      tabw[j] |= (USHORT)((j & (1 << k)) << k);

  charptr = (UCHAR *) tabw;		/* Check that Z's lookup will work */
  if (charptr[2] EQ '\0') {
    printf("Sorry, not a little-endian machine, "); 
    printf("set TABLEZ to FALSE and recompile\n");
    return(ERROR);
    }

#elif TABLE

  for (j = 255; j GE 0; j--) 		/* Set up ULONG lookup table tabw */
    for (tabw[j] = 0, k = 7; k GE 0; k--)
      tabw[j] |= (ULONG)((j & (1 << k)) << k);

#endif

  if (argc GT 4) {
    fp = myfopen(argv[4], "r");
    while (fgets(line, MC, fp) NE NULL) {
      skiprec = (struct skip *)mymalloc((int)sizeof(struct skip));
      slow = shigh = 0;  
      sscanf(line, "%d %d", &slow, &shigh);
      if ((slow LE 0) OR (slow GE shigh)) break;
#if VERBOSE
      printf("Will skip s in [%d, %d]\n", slow, shigh);
#endif      
      skiprec->low = slow; 
      skiprec->high = shigh;
      skiprec->next = skiplist;
      skiplist = skiprec;
      }
    fclose(fp);	
    }

  if (argc GT 3)
    sscanf(argv[3], "%d", &minutes);	  /* Run-time bound if nonzero,
    					     ignored if zero */
  if (argc GT 1) {
    fp = myfopen(argv[1], "r");
    fgets(line, MC, fp);   
    fclose(fp);	
    }
  else {
    printf ("r, s1, s2 ?\n"); 	  	  /* Interactive input */
    fgets(line, MC, stdin);
    }
  sscanf(line, "%d %d %d", &r, &s1, &s2);
    					  /* Searches up from initial s1
    					     to s2-1 (or down to s2+1) */ 
  if (s1 LT 0) return(OK);
  if (s1 GE r) s1 = r/2;		  /* Reasonable defaults */
  if (s2 LT 0) s2 = 0;
  if (s2 GT r) s2 = (s1 GT r/2) ? r - 64: r/2 + 1;  

  /* Swan's theorem says trinomial is reducible if r = +- 3 (mod 8)
     and s is not 2 or r-2 */
     
  swan = ((r%8) EQ 3) OR ((r%8) EQ 5);
  if (swan) {
    printf("Swan's theorem applies, only need to test s = 2\n");
    s1 = 2;
    s2 = 1;
    }   

  if (s1 LT s2)
    printf("Searching from %d to %d\n", s1, s2-1);
  else if (s1 GT s2)
    printf("Searching from %d down to %d\n", s1, s2+1);
  else 
    printf("Empty search interval at %d\n", s1);        
  if (minutes NE 0) printf("Time limit %d minutes\n", minutes);
  fflush(stdout);  
  while ((s1 NE s2) AND (NOT found) AND 
    ((minutes EQ 0) OR (CPUtotal < 60*minutes))) {

    done = FALSE;
    for (;;) {		/* Look for next s in range but not in skiplist */
      s = s1; 
      if (s1 EQ s2) {
        done = TRUE;
        break;
        }
      if (s1 LT s2)
        s1++;
      else
        s1--;
      if (NOT skips(skiplist, s)) break;		  
      }   
    if (done) break;  
      
#if VERBOSE       
    printf ("\nr %d, s %d, r-s %d\n", r, s, r-s);
#endif
    if ((s LE 0) OR (r LE s) OR ((r-s) LT WLEN2)) {
      printf ("Illegal parameters\n");
      return(ERROR);
      }

    CPUtime += clockd(&cstart, FALSE);
    CPUtotal += CPUtime;		/* CPUtotal is for all (r, s) */
    CPUtime = 0;			/* CPUtime is for current (r, s) */
    CPUtime1 = 0;			/* CPUtime1 is for current sieving */
    CPUlast = 0;
    skt++;

    /* Set up global variables depending only on r and s */

    q1  = r >> WD;			/* q1 GE q2 > 0 */
    r1  = r & WLENM;			/* 0 LE r1 < WLEN */
    r1c = WLEN - r1;			/* Could be WLEN */
    q2  = (r-s) >> WD;			/* q2 > 0 (since r-s GE WLEN) */
    r2  = (r-s) & WLENM;		/* 0 LE r2 < WLEN */
    r2c = WLEN - r2;			/* Could be WLEN */

    sizep = q1 + 3;			/* sizep = r/WLEN + 3 (this fixes  */
    sizea = 2*sizep;			/* bug in versions 1.30, 1.31) */

    /* Sieve to discard polynomials which are easily found to be reducible */

    kn = 1;
    for (nm = n = 2; n LE (NMAX-NEXTRA); n++) { /* Don't need to consider 
    						   n = 1 for a trinomial */
      kn = 2*kn + 1;			/* kn = 2^n - 1 */
      if (kn GE r) break;		/* choose nm so 2^nm < r */
      nm = n;
      }
      
    if (a EQ NULL) {	/* Allocate space for (a, p) and q in one block.
      			   The "+ 4" allows for increasing a by 2 or 3. */
      			   		
      a = (ULONG *)mymalloc((sizea + sizep + 4)*(int)sizeof(ULONG));
	
      a += 2;		/* To simplify optimisations in versions of square, 
        		   we want to access a[-1], a[-2]. */
        		     
      /* Following is unnecessary if malloc always allocates in 8-byte
	   chunks, but do it just in case, to align a[0] on 8-byte boundary */

#if (NOT W64)
      if ((((UINT)a) & 4) NE 0) a += 1;
#endif
     
      p = a + (sizea - sizep);	/* Can overlap p with top half of a */
      q = p + sizep;
      }  

    kn = 1;
    g = TRUE;
    for (n = 2; n LE nm; n++) {	/* n EQ 1 never succeeds for trinomial, */
      nsv = n;			/* for n > 1 prob. of success about 1/(n+1) */
      kn = 2*kn + 1;		/* 2^n - 1 */
      for (j = kn >> WD; j GE 0; j--)
        p[j] = 0;
      p[0] = 1;
      p[kn >> WD] ^= 1L << (kn & WLENM);	/* p = x^kn + 1 */
      pn = kn;
      rk = r%kn;
      sk = s%kn;
      /* Following is intended to speed up relprime */
      if (rk < sk){			/* ensure that rk = max (rk, sk) */
        qn = rk; rk = sk; sk = qn;
        }
      if (rk < 2*sk) sk = rk - sk;	/* so now 0 LE 2*sk LE rk */
      qn = rk;	  			/* degree of q (even if rk = sk = 0) */
      for (j = qn >> WD; j GE 0; j--)
        q[j] = 0;
      q[0] = 1;
      q[rk >> WD] ^= 1L << (rk & WLENM);
      q[sk >> WD] ^= 1L << (sk & WLENM); /* q = x^rk + x^sk + 1 */
      g = relprime(p, pn, q, qn);
      if (NOT g) break;
      CPUtime += clockd(&cstart, FALSE);
      }

    if (NOT g) {
      kt2++;
#if VERBOSE      
      printf("Reducible by test with n %d\n", nsv);
      fflush(stdout);
#endif      
      sieved[nsv]++;
      if (argc GT 2) {
        fp = myfopen(argv[2], "a");
        fprintf(fp, "%d %d %d\n", r, s, nsv);
        fclose(fp);
        }
      else {
        printf("%d %d %d%s\n", r, s, nsv, log);
        fflush(stdout);
        }
      }

    CPUtime += clockd(&cstart, FALSE);

    if (g) { /* Phase 1 sieve did not decide reducibility */

      setupx (a);
      for (k = 0; g AND (k < r); k++) {
#if (IBMPC AND ASM)
 	squarem(a, q1);			/* Use assembler routine */
#elif TABLE
	squaret (a, tabw, q1);		/* Use table lookup */
#else      
	square (a);			/* Use one of four other versions */
#endif
        reduce (a);
        if ((k & 0x7FFFL) EQ 0) { 	 	/* Avoid timer overflow by */
          CPUtime += clockd(&cstart, FALSE);	/* calling clockd sometimes */
          }

      /* Phase 2 sieve: compute GCD(a + x), trinomial) for small k */
         
        if ((k GE nm) AND (k LT (nm+NEXTRA)) AND (k LT (r-1))) {
	  for (j = q1; j GE 0; j--) 
            p[j] = a[j];		/* Copy as relprime destroys p, q */
	  p[0] ^= 2L;			/* Add x (mod 2) */
	  pn = r;			/* Upper bound on degree */
	  for (j = pn; j GE 0; j--) {	/* Find exact degree */
	    new = p[j >> WD];
      	    if (new EQ 0L) {		/* Speed up search over zero words */
	      pn -= j & WLENM;
 	      j-= j & WLENM;		
	      }
	    else
	      if ((new >> (j & WLENM)) &1L) break;
      	    pn--;
	    }
	  qn = r;
	  for (j = qn >> WD; j GE 0; j--) q[j] = 0;
	  q[0] = 1;
	  q[r >> WD] ^= 1L << (r & WLENM);
	  q[s >> WD] ^= 1L << (s & WLENM); /* q = x^r + x^s + 1 */

	  g = relprime(p, pn, q, qn);

#if VERBOSE
	  if (k LT (nm + NEXTRA)) {
	    CPUtime += clockd(&cstart, FALSE);
	    CPUlast = CPUtime;
	    CPUtime1 += CPUtime;	/* CPUtime1 is for sieving */
	    CPUtotal += CPUtime;
	    CPUtime = 0;		/* Non-sieving from now on */
	    }
#endif
	  if (g) {
	    if (k EQ (nm + (NEXTRA - 1))) { 		
	      kt1++;			/* Unsuccessful sieve count */
	      sieved[0]++;
	      }
            }
          else {
	    kt2++;
#if VERBOSE      
	    printf("Reducible by test with n %d\n", k+1);
#endif      
	    sieved[k+1]++;
	    if (argc GT 2) {
	      fp = myfopen(argv[2], "a");
	      fprintf(fp, "%d %d %d\n", r, s, k+1);
	      fclose(fp);
	      }
	    else {
	      printf("%d %d %d%s\n", r, s, k+1, log); /* bug in 1.31 fixed */
	      fflush(stdout);
	      }
	    }
	  }
#if VERBOSE	    
	if ((k EQ (nm + NEXTRA)) AND (CPUtime1 GT CPUtol)) {
	  printf("Sieving unsuccessful after %.2f sec", CPUtime1); 
	  if (CPUlast GT CPUtol) 
	    printf(", degree %d took %.2f sec", k, CPUlast);
	  printf("\n");
	  fflush(stdout);
	  }
#endif	      
	if (NOT g) break;  
        }

      CPUtime += clockd(&cstart, FALSE);

      if (g) {       
       if (comparex(a)) {
        found = (NOT CONTINUE);
        printf("%d %d primitive (if 2^%d-1 is prime)\n", r, s, r);
        fflush(stdout);
        if (argc GT 2) {
          fp = myfopen(argv[2], "a");
	  fprintf(fp, "%d %d primitive\n", r, s);
 	  fclose(fp);
          }
        else {
	  printf("%d %d primitive%s\n", r, s, log);
	  fflush(stdout);
	  }
        }
      else {
#if W64
	temp = (int)((a[0]<<32)>>32);
#else      
        temp = (int)a[0];
#endif
#if VERBOSE
        printf("Not primitive, low word %8x (hex)\n", temp); 
#endif
        if (argc GT 2) {
          fp = myfopen(argv[2], "a");
 	  fprintf(fp, "%d %d x%8x\n", r, s, temp);
 	  fclose(fp);
          }
        else {
 	  printf("%d %d x%8x%s\n", r, s, temp, log);
 	  fflush(stdout);
 	  }
        }
      }
      
#if VERBOSE        
      if (CPUtime GT CPUtol) {  
        printf("CPU time (not sieving) %.2f sec = ", CPUtime);
        printf("%2.2f r^2 nsec\n", 1.0e9*CPUtime/r/r);
        }
      if (CPUtime1 GT CPUtol) {
	printf("CPU time (sieving)     %.2f sec", CPUtime1);
	if (CPUtime GT CPUtime1)
          printf(", ratio %2.0f", CPUtime/CPUtime1);
        printf("\n");   
        }
      CPUtotal1 += CPUtime1;  
      CPUtotal += CPUtime;
      CPUtime = 0;
      if (CPUtotal GT CPUtol) {
        printf("\nOverall time %.2f sec = ", CPUtotal);
        printf("%.4f sec per trinomial\n", CPUtotal/skt);
        if ((CPUtotal1 GT CPUtol) AND (CPUtotal1 LT CPUtotal))
          printf("Overall %1.2f percent of time spent sieving\n", 
        	100.0*CPUtotal1/CPUtotal); 
	printf("Sieving: success %d, failure %d, ", kt2, kt1);
	printf("rate %1.2f percent\n", 100.0*(double)kt1/(kt1+kt2));
	for (j = 1; j LE (nm+NEXTRA); j++) printf ("%d ", sieved[j]);
	printf("\n");
        }
      fflush(stdout);  
#endif        
      }
  
    /* Try again until primitive trinomial found
       (or until all s checked if CONTINUE is true) */

    if (argc GT 1) {
      fp = myfopen(argv[1], "w");
      fprintf(fp, "%d %d %d\n", r, s1, s2);
      fclose(fp);
      }  

    if (swan) break;
    }

  if (argc GT 1) {
    fp = myfopen(argv[1], "w");
    fprintf(fp, "%d %d %d\n", r, s1, s2);
    fclose(fp);
    }

  if ((NOT found) AND (p NE NULL)) 
    printf("Searched to %d\n", s);
#if VERBOSE
  printf("\nDynamic storage allocation %d bytes\n", space);
#endif  
  return(OK);				/* Normal exit */
  }
